/*!\file: CreateParametersQmu.cpp
 * \brief general driver for creating parameters dataset
 */ 

#include "../../../Container/Container.h"
#include "../../../toolkits/toolkits.h"
#include "../../../EnumDefinitions/EnumDefinitions.h"
#include "../../../objects/objects.h"
#include "../../../shared/shared.h"
#include "../../../io/io.h"
#include "../../../include/include.h"
#include "../../MeshPartitionx/MeshPartitionx.h"
#include "../ModelProcessorx.h"

void CreateParametersQmu(Parameters** pparameters,IoModel* iomodel,ConstDataHandle iomodel_handle,int analysis_type){

	/*variable declarations: {{{1*/
	int i,j,k;
	
	Parameters* parameters = NULL;
	int      second_count;
	
	int*     part=NULL;
	double*  dpart=NULL;

	char**  responsedescriptors=NULL;
	char**  variabledescriptors=NULL;
	char*   descriptor=NULL;
	double* dakota_parameter=NULL;

	//qmu files
	char* qmuinname=NULL;
	char* qmuerrname=NULL;
	char* qmuoutname=NULL;
	extern int my_rank;
	
	//tags
	char variabledescriptors_tag[30]; //strlen("variabledescriptor"+number)+1
	char responsedescriptors_tag[30]; //strlen("responsedescriptors"+number)+1
	char massflux_tag[35];            //strlen("qmu_mass_flux_segments"+number)+1
	
	/*parameters for mass flux: */
	bool     qmu_mass_flux_present=false;
	double** array=NULL;
	double*  matrix=NULL;
	int*     mdims_array=NULL;
	int*     ndims_array=NULL;
	int      M;

	#ifdef _SERIAL_
		mxArray* pfield=NULL;
		mxArray* pfield2=NULL;
	#endif
	/*}}}*/
	
	/*recover parameters : */
	parameters=*pparameters;

	parameters->AddObject(new   BoolParam(QmuAnalysisEnum,iomodel->qmu_analysis));
	if(iomodel->qmu_analysis){

		/*name of qmu input, error and output files:{{{1*/
		qmuinname=(char*)xmalloc((strlen(iomodel->name)+strlen(".qmu.in")+1)*sizeof(char));
		sprintf(qmuinname,"%s%s",iomodel->name,".qmu.in");
		parameters->AddObject(new   StringParam(QmuInNameEnum,qmuinname));

		qmuoutname=(char*)xmalloc((strlen(iomodel->name)+strlen(".qmu.out")+1)*sizeof(char));
		sprintf(qmuoutname,"%s%s",iomodel->name,".qmu.out");
		parameters->AddObject(new   StringParam(QmuOutNameEnum,qmuoutname));

		qmuerrname=(char*)xmalloc((strlen(iomodel->name)+strlen(".qmu.err")+1)*sizeof(char));
		sprintf(qmuerrname,"%s%s",iomodel->name,".qmu.err");
		parameters->AddObject(new   StringParam(QmuErrNameEnum,qmuerrname));
		/*}}}*/
		
		/*Deal with variable descriptors: {{{1*/
		variabledescriptors=(char**)xmalloc(iomodel->numberofvariables*sizeof(char*));


		/*Fetch descriptors: logic varies if we are running parallel or serial. In parallel, qmumarshall 
		 * took care of marshalling all the variable descriptors, so it's easy. In serial mode, 
		 * the variables are in md.variables(md.ivar), as a strucuture: */

		#ifdef _SERIAL_
		pfield=mxGetField(iomodel_handle,0,"variabledescriptors");
		for(i=0;i<iomodel->numberofvariables;i++){
			pfield2=mxGetCell(pfield,i);
			FetchData(&descriptor,pfield2);
			variabledescriptors[i]=descriptor;
		}
		#else
		for(i=0;i<iomodel->numberofvariables;i++){
			sprintf(variabledescriptors_tag,"%s%i","variabledescriptor",i);
			IoModelFetchData(&descriptor,iomodel_handle,variabledescriptors_tag);
			variabledescriptors[i]=descriptor;
		}
		#endif

		/*Ok, we have all the variable descriptors. Build a parameter with it: */
		parameters->AddObject(new StringArrayParam(VariableDescriptorsEnum,variabledescriptors,iomodel->numberofvariables));

		/*}}}*/

		/*Deal with responses: {{{1*/
		responsedescriptors=(char**)xmalloc(iomodel->numberofresponses*sizeof(char*));

		/*Fetch descriptors: */
		#ifdef _SERIAL_
		pfield=mxGetField(iomodel_handle,0,"responsedescriptors");
		for(i=0;i<iomodel->numberofresponses;i++){
			pfield2=mxGetCell(pfield,i);
			FetchData(&descriptor,pfield2);
			responsedescriptors[i]=descriptor;
		}
		#else
		for(i=0;i<iomodel->numberofresponses;i++){
			sprintf(responsedescriptors_tag,"%s%i","responsedescriptor",i);
			IoModelFetchData(&descriptor,iomodel_handle,responsedescriptors_tag);
			responsedescriptors[i]=descriptor;
		}
		#endif

		/*Ok, we have all the response descriptors. Build a parameter with it: */
		parameters->AddObject(new StringArrayParam(ResponseDescriptorsEnum,responsedescriptors,iomodel->numberofresponses));
		/*}}}}*/

		/*Deal with partitioning: {{{1*/
		/*partition vertices in iomodel->qmu_npart parts, unless a partition is already present: */
		parameters->AddObject(new    IntParam(QmuNPartEnum,iomodel->qmu_npart));
		
		IoModelFetchData(&dpart,NULL,NULL,iomodel_handle,"part");

		if(!dpart){

			/*Partition elements and vertices and nodes: */
			ElementsAndVerticesPartitioning(&iomodel->my_elements,&iomodel->my_vertices, &iomodel->my_bordervertices, iomodel,iomodel_handle);

			dpart=(double*)xmalloc(iomodel->numberofvertices*sizeof(double));
			for(i=0;i<iomodel->numberofvertices;i++)dpart[i]=iomodel->my_vertices[i];
		}
		parameters->AddObject(new DoubleVecParam(QmuPartEnum,dpart,iomodel->numberofvertices));
		/*}}}*/

		/*Deal with data needed because of qmu variable inputs: {{{1*/
		for(i=0;i<iomodel->numberofvariables;i++){

			descriptor=variabledescriptors[i];

			if ((strcmp(descriptor,"Thickness")==0) ||
				(strcmp(descriptor,"DragCoefficient")     ==0)
				){

				IoModelFetchData(&dakota_parameter,NULL,NULL,iomodel_handle,EnumAsModelField(StringAsEnum(descriptor)));
				parameters->AddObject(new DoubleVecParam(StringAsEnum(descriptor),dakota_parameter,iomodel->numberofvertices));
				xfree((void**)&dakota_parameter);
			}
		}
		/*}}}*/

		/*Deal with data needed to compute qmu responses: {{{1*/
		for(i=0;i<iomodel->numberofresponses;i++){
			char* descriptor=responsedescriptors[i];
			
			if(strncmp(descriptor,"MassFlux",8)==0){
				qmu_mass_flux_present=true;
			}

			if (strcmp(descriptor,"misfit")==0){

				/*We need the fit: */
				IoModelFetchData(&iomodel->fit,NULL,NULL,iomodel_handle,"fit");
				parameters->SetParam(iomodel->fit,iomodel->nsteps,FitEnum);
				xfree((void**)&iomodel->fit);

			}
		}
		
		
		if(qmu_mass_flux_present){
			
			/*We need the segments to be able to compute the mass flux. We have as many groups of segments as we have MassFlux 
			 *responses. Let's build a DoubleMatArrayParam object with the array of segments: */
			
			if(iomodel->qmu_mass_flux_num_profiles==0)ISSMERROR(" qmu_mass_flux_num_profiles is 0, when MassFlux computations were requested!");
			array=(double**)xmalloc(iomodel->qmu_mass_flux_num_profiles*sizeof(double*));
			mdims_array=(int*)xmalloc(iomodel->qmu_mass_flux_num_profiles*sizeof(int));
			ndims_array=(int*)xmalloc(iomodel->qmu_mass_flux_num_profiles*sizeof(int));
			M=iomodel->qmu_mass_flux_num_profiles;

			#ifdef _SERIAL_
			pfield=mxGetField(iomodel_handle,0,"qmu_mass_flux_segments");
			for(i=0;i<iomodel->qmu_mass_flux_num_profiles;i++){
				pfield2=mxGetCell(pfield,i);
				FetchData(&matrix,mdims_array+i,ndims_array+i,pfield2);
				array[i]=matrix;
			}
			#else
			for(i=0;i<iomodel->qmu_mass_flux_num_profiles;i++){
				sprintf(massflux_tag,"%s%i","qmu_mass_flux_segmentsi",i);
				IoModelFetchData(&tot_matrix,tot_mdims_array+i,NULL,iomodel_handle,"qmu_mass_flux_segments");
				variabledescriptors[i]=descriptor;
			}
			#endif

			#ifdef _PARALLEL_
			IoModelFetchData(&qmu_mass_flux_segments,&tot_qmu_mass_flux_segments,NULL,iomodel_handle,"qmu_mass_flux_segments");
			IoModelFetchData(&qmu_mass_flux_num_segments,&tot_qmu_mass_flux_num_segments,NULL,iomodel_handle,"qmu_mass_flux_num_segments");


			/*Only if partitioning exist do we care about the segments: */
			if(iomodel->my_elements){

				/*Use the element partitioning vector from the iomodel to down select qmu_mass_flux_segments to only segments that are relevant 
				 * to this cpu: */
				my_tot_qmu_mass_flux_segments=0;
				for(j=0;j<tot_qmu_mass_flux_segments;j++){
					if (  iomodel->my_elements[(int)(*(qmu_mass_flux_segments+5*j+4))-1])my_tot_qmu_mass_flux_segments++;
				}


				if(my_tot_qmu_mass_flux_segments){
					my_qmu_mass_flux_segments=(double*)xcalloc(5*my_tot_qmu_mass_flux_segments,sizeof(double));
					second_count=0;
					for(j=0;j<tot_qmu_mass_flux_segments;j++){
						if (iomodel->my_elements[(int)*(qmu_mass_flux_segments+5*j+4)-1]){
							for(k=0;k<5;k++)*(my_qmu_mass_flux_segments+5*second_count+k)=*(qmu_mass_flux_segments+5*j+k);
							second_count++;
						}
					}
				}

				parameters->AddObject(new DoubleMatParam(QmuMassFluxSegmentsEnum,my_qmu_mass_flux_segments,my_tot_qmu_mass_flux_segments,5));

			}
					
			#else
			IoModelFetchData(&qmu_mass_flux_segments,&tot_qmu_mass_flux_segments,NULL,iomodel_handle,"qmu_mass_flux_segments");
			IoModelFetchData(&qmu_mass_flux_num_segments,&tot_qmu_mass_flux_num_segments,NULL,iomodel_handle,"qmu_mass_flux_num_segments");

			parameters->AddObject(new DoubleMatParam(QmuMassFluxSegmentsEnum,qmu_mass_flux_segments,tot_qmu_mass_flux_segments,5));
			parameters->AddObject(new DoubleVecParam(QmuMassFluxNumSegmentsEnum,qmu_mass_flux_num_segments,tot_qmu_mass_flux_num_segments));

			#endif

			xfree((void**)&qmu_mass_flux_segments);
			xfree((void**)&my_qmu_mass_flux_segments);
			xfree((void**)&qmu_mass_flux_num_segments);
			xfree((void**)&my_qmu_mass_flux_num_segments);
		}
		/*}}}*/

		/*Free data: {{{1*/
		for(i=0;i<iomodel->numberofresponses;i++){
			char* descriptor=responsedescriptors[i];
			xfree((void**)&descriptor);
		}
		xfree((void**)&responsedescriptors);

		for(i=0;i<iomodel->numberofvariables;i++){
			char* descriptor=variabledescriptors[i];
			xfree((void**)&descriptor);
		}
		xfree((void**)&variabledescriptors);

		xfree((void**)&iomodel->elements);
		xfree((void**)&iomodel->elements2d);
		xfree((void**)&part);
		xfree((void**)&dpart);
		xfree((void**)&qmuinname);
		xfree((void**)&qmuerrname);
		xfree((void**)&qmuoutname);
		/*}}}*/
	} //if(iomodel->qmu_analysis)

	/*Assign output pointer: */
	*pparameters=parameters;
}
